import random
from uuid import uuid4

import pytest
from orchestrator.db import db
from orchestrator.domain import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle, UUIDstr

from gso.products import ProductName
from gso.products.product_blocks.bgp_session import BGPSession, IPFamily, IPTypes
from gso.products.product_blocks.l3_core_service import AccessPort
from gso.products.product_blocks.service_binding_port import BFDSettings, ServiceBindingPort
from gso.products.product_types.edge_port import EdgePort
from gso.products.product_types.l3_core_service import (
    ImportedL3CoreService,
    L3CoreServiceInactive,
    L3CoreServiceType,
)
from gso.services import subscriptions
from gso.utils.shared_enums import APType, SBPType
from gso.utils.types.ip_address import IPAddress


@pytest.fixture()
def bfd_settings_factory(faker):
    def create_bfd_settings(
        bfd_multiplier: int | None = None,
        bfd_interval: int | None = None,
        *,
        bfd_enabled: bool | None = None,
    ):
        bfd_enabled = bfd_enabled or faker.boolean()
        return BFDSettings.new(
            subscription_id=uuid4(),
            bfd_enabled=bfd_enabled,
            bfd_multiplier=bfd_multiplier or (faker.pyint() if bfd_enabled else None),
            bfd_interval_rx=bfd_interval or (faker.pyint() if bfd_enabled else None),
            bfd_interval_tx=bfd_interval or (faker.pyint() if bfd_enabled else None),
        )

    return create_bfd_settings


@pytest.fixture()
def bgp_session_subscription_factory(faker):
    def create_bgp_session(
        peer_address: IPAddress | None = None,
        families: list[IPFamily] | None = None,
        authentication_key: str | None = None,
        *,
        is_multi_hop: bool = False,
        has_custom_policies: bool = False,
        multipath_enabled: bool | None = True,
        send_default_route: bool | None = True,
        is_passive: bool | None = False,
        rtbh_enabled: bool | None = False,
        bfd_enabled: bool | None = False,
        ip_type: IPTypes | None = None,
    ):
        return BGPSession.new(
            subscription_id=uuid4(),
            peer_address=peer_address or faker.ipv4(),
            bfd_enabled=bfd_enabled if bfd_enabled else faker.boolean(),
            families=families or [IPFamily.V4UNICAST],
            has_custom_policies=has_custom_policies,
            authentication_key=authentication_key or faker.password(),
            multipath_enabled=multipath_enabled,
            send_default_route=send_default_route,
            is_multi_hop=is_multi_hop,
            rtbh_enabled=rtbh_enabled,
            is_passive=is_passive,
            ip_type=ip_type if ip_type else faker.random_element(elements=[IPTypes.IPV4, IPTypes.IPV6]),
        )

    return create_bgp_session


@pytest.fixture()
def service_binding_port_factory(
    faker, bgp_session_subscription_factory, edge_port_subscription_factory, bfd_settings_factory
):
    def create_service_binding_port(
        bgp_session_list: list[BGPSession] | None = None,
        geant_sid: str | None = None,
        sbp_type: SBPType = SBPType.L3,
        ipv4_address: str | None = None,
        ipv4_mask: int | None = None,
        ipv6_address: str | None = None,
        ipv6_mask: int | None = None,
        vlan_id: int | None = None,
        edge_port: EdgePort | None = None,
        v4_bfd_settings: BFDSettings | None = None,
        v6_bfd_settings: BFDSettings | None = None,
        *,
        custom_firewall_filters: bool = False,
        is_tagged: bool = False,
    ):
        return ServiceBindingPort.new(
            subscription_id=uuid4(),
            is_tagged=is_tagged,
            vlan_id=vlan_id or faker.vlan_id(),
            sbp_type=sbp_type,
            ipv4_address=ipv4_address or faker.ipv4(),
            ipv4_mask=ipv4_mask or faker.ipv4_netmask(),
            ipv6_address=ipv6_address or faker.ipv6(),
            ipv6_mask=ipv6_mask or faker.ipv6_netmask(),
            custom_firewall_filters=custom_firewall_filters,
            geant_sid=geant_sid or faker.geant_sid(),
            bgp_session_list=bgp_session_list
            or [
                bgp_session_subscription_factory(families=[IPFamily.V4UNICAST]),
                bgp_session_subscription_factory(families=[IPFamily.V6UNICAST], peer_address=faker.ipv6()),
            ],
            edge_port=edge_port or EdgePort.from_subscription(edge_port_subscription_factory()).edge_port,
            v4_bfd_settings=v4_bfd_settings or bfd_settings_factory(),
            v6_bfd_settings=v6_bfd_settings or bfd_settings_factory(),
        )

    return create_service_binding_port


@pytest.fixture()
def access_port_factory(faker, service_binding_port_factory):
    def create_access_port(
        ap_type: APType | None = None,
        service_binding_port: ServiceBindingPort | None = None,
    ):
        return AccessPort.new(
            subscription_id=uuid4(),
            ap_type=ap_type or random.choice(list(APType)),  # noqa: S311
            sbp=service_binding_port or service_binding_port_factory(),
        )

    return create_access_port


@pytest.fixture()
def l3_core_service_subscription_factory(
    faker,
    partner_factory,
    access_port_factory,
):
    def create_l3_core_service_subscription(
        l3_core_service_type: L3CoreServiceType,
        description=None,
        partner: dict | None = None,
        ap_list: list[AccessPort] | None = None,
        start_date="2023-05-24T00:00:00+00:00",
        status: SubscriptionLifecycle | None = None,
    ) -> UUIDstr:
        partner = partner or partner_factory()
        match l3_core_service_type:
            case L3CoreServiceType.GEANT_IP:
                product_id = subscriptions.get_product_id_by_name(ProductName.GEANT_IP)
                l3_core_service_subscription = L3CoreServiceInactive.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.IMPORTED_GEANT_IP:
                product_id = subscriptions.get_product_id_by_name(ProductName.IMPORTED_GEANT_IP)
                l3_core_service_subscription = ImportedL3CoreService.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.IAS:
                product_id = subscriptions.get_product_id_by_name(ProductName.IAS)
                l3_core_service_subscription = L3CoreServiceInactive.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.IMPORTED_IAS:
                product_id = subscriptions.get_product_id_by_name(ProductName.IMPORTED_IAS)
                l3_core_service_subscription = ImportedL3CoreService.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.GWS:
                product_id = subscriptions.get_product_id_by_name(ProductName.GWS)
                l3_core_service_subscription = L3CoreServiceInactive.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.IMPORTED_GWS:
                product_id = subscriptions.get_product_id_by_name(ProductName.IMPORTED_GWS)
                l3_core_service_subscription = ImportedL3CoreService.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.LHCONE:
                product_id = subscriptions.get_product_id_by_name(ProductName.LHCONE)
                l3_core_service_subscription = L3CoreServiceInactive.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.IMPORTED_LHCONE:
                product_id = subscriptions.get_product_id_by_name(ProductName.IMPORTED_LHCONE)
                l3_core_service_subscription = ImportedL3CoreService.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.COPERNICUS:
                product_id = subscriptions.get_product_id_by_name(ProductName.COPERNICUS)
                l3_core_service_subscription = L3CoreServiceInactive.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case L3CoreServiceType.IMPORTED_COPERNICUS:
                product_id = subscriptions.get_product_id_by_name(ProductName.IMPORTED_COPERNICUS)
                l3_core_service_subscription = ImportedL3CoreService.from_product_id(
                    product_id, customer_id=partner["partner_id"], insync=True
                )
            case _:
                msg = f"L3 Core Service type not found: {l3_core_service_type}"
                raise ValueError(msg)

        # Default ap_list creation with primary and backup access ports
        l3_core_service_subscription.l3_core_service.ap_list = ap_list or [
            access_port_factory(ap_type=APType.PRIMARY),
            access_port_factory(ap_type=APType.BACKUP),
        ]

        # Update subscription with description, start date, and status
        l3_core_service_subscription = SubscriptionModel.from_other_lifecycle(
            l3_core_service_subscription,
            SubscriptionLifecycle.ACTIVE,
        )
        l3_core_service_subscription.description = description or faker.sentence()
        l3_core_service_subscription.start_date = start_date
        l3_core_service_subscription.status = status or SubscriptionLifecycle.ACTIVE
        l3_core_service_subscription.save()

        db.session.commit()

        return str(l3_core_service_subscription.subscription_id)

    return create_l3_core_service_subscription
